r"""
Interval exchange transformations

This library is designed for the usage and manipulation of interval
exchange transformations and linear involutions. It defines specialized
types of permutation (constructed using :func:`Permutation`) some
associated graph (constructed using :func:`iet.RauzyGraph`) and some maps
of intervals (constructed using :func:`IntervalExchangeTransformation`).


EXAMPLES::

    sage: from surface_dynamics import *

Creation of an interval exchange transformation (iet)::

    sage: T = iet.IntervalExchangeTransformation(('a b','b a'),(sqrt(2),1))
    sage: T
    Interval exchange transformation of [0, sqrt(2) + 1[ with permutation
    a b
    b a

It can also be initialized using permutation (group theoritic ones)::

    sage: p = Permutation([3,2,1])
    sage: T = iet.IntervalExchangeTransformation(p, [1/3,2/3,1])
    sage: T
    Interval exchange transformation of [0, 2[ with permutation
    1 2 3
    3 2 1


As the iet's are functions, you can compose and invert them::

    sage: T = iet.IntervalExchangeTransformation(('a b','b a'),(sqrt(2),1))
    sage: T*T
    Interval exchange transformation of [0, sqrt(2) + 1[ with permutation
    aa ab ba
    ab ba aa
    sage: S = T.inverse()
    sage: S
    Interval exchange transformation of [0, sqrt(2) + 1[ with permutation
    b a
    a b
    sage: S * T
    Interval exchange transformation of [0, sqrt(2) + 1[ with permutation
    aa bb
    aa bb
    sage: (S * T).is_identity()
    True
    sage: T * S
    Interval exchange transformation of [0, sqrt(2) + 1[ with permutation
    bb aa
    bb aa
    sage: (T * S).is_identity()
    True

For the manipulation of permutations of iet, there are special types provided by
this module. All of them can be constructed using the constructor
iet.Permutation. For the creation of labelled permutations of interval exchange
transformation::

    sage: p1 =  iet.Permutation('a b c', 'c b a')
    sage: p1
    a b c
    c b a

They can be used for initialization of an iet::

    sage: p = iet.Permutation('a b', 'b a')
    sage: T = iet.IntervalExchangeTransformation(p, [1,sqrt(2)])
    sage: T
    Interval exchange transformation of [0, sqrt(2) + 1[ with permutation
    a b
    b a

You can also, create labelled permutations of linear involutions::

    sage: p = iet.GeneralizedPermutation('a a b', 'b c c')
    sage: p
    a a b
    b c c

By default, the permutations are labelled (it means that the labels are
important and (a b / b a) differs from (b a / a b)). It sometimes useful to deal
with reduced permutations for which the order does not import::

    sage: p = iet.Permutation('a b c', 'c b a', reduced = True)
    sage: p
    a b c
    c b a

Permutations with flips::

    sage: p1 = iet.Permutation('a b c', 'c b a', flips = ['a','c'])
    sage: p1
    -a  b -c
    -c  b -a

Creation of Rauzy diagrams::

    sage: r = iet.RauzyDiagram('a b c', 'c b a')

Reduced Rauzy diagrams are constructed using the same arguments than for
permutations::

    sage: r = iet.RauzyDiagram('a b b','c c a')
    sage: r_red = iet.RauzyDiagram('a b b','c c a',reduced=True)
    sage: r.cardinality()
    12
    sage: r_red.cardinality()
    4

By defaut, Rauzy diagram are generated by induction on the right. You can use
several options to enlarge (or restrict) the diagram (try help(iet.RauzyDiagram) for
more precisions)::

    sage: r1 = iet.RauzyDiagram('a b c','c b a',right_induction=True)
    sage: r2 = iet.RauzyDiagram('a b c','c b a',left_right_inversion=True)

You can consider self similar iet using path in Rauzy diagrams and eigenvectors
of the corresponding matrix::

    sage: p = iet.Permutation("a b c d", "d c b a")
    sage: d = p.rauzy_diagram()
    sage: g = d.path(p, 't', 't', 'b', 't', 'b', 'b', 't', 'b')
    sage: g
    Path of length 8 in a Rauzy diagram
    sage: g.is_loop()
    True
    sage: g.is_full()
    True
    sage: m = g.matrix()
    sage: v = m.eigenvectors_right()[-1][1][0]
    sage: T1 = iet.IntervalExchangeTransformation(p, v)
    sage: T2 = T1.rauzy_move(iterations=8)
    sage: T1.normalize(1) == T2.normalize(1)
    True

REFERENCES:

.. [BL08] Corentin Boissy and Erwan Lanneau, "Dynamics and geometry of the
  Rauzy-Veech induction for quadratic differentials" (arxiv:0710.5614) to appear
  in Ergodic Theory and Dynamical Systems

.. [DN90] Claude Danthony and Arnaldo Nogueira "Measured foliations on
  nonorientable surfaces", Annales scientifiques de l'Ecole Normale
  Superieure, Ser. 4, 23, no. 3 (1990) p 469-494

.. [N85] Arnaldo Nogueira, "Almost all Interval Exchange Transformations with
  Flips are Nonergodic" (Ergod. Th. & Dyn. Systems, Vol 5., (1985), 257-271

.. [R79] Gerard Rauzy, "Echanges d'intervalles et transformations induites",
  Acta Arith. 34, no. 3, 203-212, 1980

.. [V78] William Veech, "Interval exchange transformations", J. Analyse Math.
  33, 222-272

.. [Z] Anton Zorich, "Generalized Permutation software"
  (http://perso.univ-rennes1.fr/anton.zorich)


AUTHORS:

- Vincent Delecroix (2009-09-29): initial version

"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

from sage.rings.integer import Integer

def _two_lists(arg1, arg2):
    r"""
    Try to return the input as a list of two lists

    INPUT:

    - ``a`` - either a string, one or two lists, one or two tuples

    OUTPUT:

    -- two lists

    TESTS::

        sage: from surface_dynamics.interval_exchanges.constructors import _two_lists

        sage: _two_lists(('a1 a2','b1 b2'), None)
        [['a1', 'a2'], ['b1', 'b2']]
        sage: _two_lists('a1 a2\nb1 b2', None)
        [['a1', 'a2'], ['b1', 'b2']]
        sage: _two_lists(['a b','c'], None)
        [['a', 'b'], ['c']]

    Make sure that tuples are properly converted::

        sage: _two_lists(((0, 1), (1, 0)), None)
        [[0, 1], [1, 0]]
        sage: _two_lists((0, 1), (1, 0))
        [[0, 1], [1, 0]]

    ValueError or TypeError are raised with invalid arguments::

        sage: _two_lists('a b', None)
        Traceback (most recent call last):
        ...
        ValueError: the chain must contain two lines
        sage: _two_lists('a b\nc d\ne f', None)
        Traceback (most recent call last):
        ...
        ValueError: the chain must contain two lines
        sage: _two_lists(1, 1)
        Traceback (most recent call last):
        ...
        TypeError: argument not accepted
    """
    from sage.combinat.permutation import Permutation as CombinatPermutation
    from sage.groups.perm_gps.permgroup_element import PermutationGroupElement

    if arg2 is None:
        if isinstance(arg1, str):
            t = arg1.split('\n')
            if len(t) != 2:
                raise ValueError("the chain must contain two lines")
            return [t[0].split(), t[1].split()]

        elif isinstance(arg1, CombinatPermutation):
            return [list(range(1,len(arg1)+1)), list(arg1)]

        elif isinstance(arg1, PermutationGroupElement):
            dom = list(arg1.parent().domain())
            return [dom, [arg1(i) for i in dom]]

        elif isinstance(arg1, (tuple,list)):
            try:
                t = CombinatPermutation(arg1)
            except Exception:
                if len(arg1) != 2:
                    raise ValueError('argument not accepted')
                arg1, arg2 = arg1
            else:
                return [list(range(1, len(t) + 1)), list(t)]

    if arg2 is None:
        raise ValueError("argument can not be split into two parts")

    res = []
    for a in (arg1, arg2):
        if isinstance(a, (tuple, list)):
            res.append(list(a))
        elif isinstance(a, str):
            res.append(a.split())
        else:
            raise TypeError('argument not accepted')

    return res

def Permutation(arg1, arg2=None, reduced=None, flips=None, alphabet=None):
    r"""
    Returns a permutation of an interval exchange transformation.

    Those permutations are the combinatoric part of an interval exchange
    transformation (IET). The combinatorial study of those objects starts with
    Gerard Rauzy [R79]_ and William Veech [V78]_.

    The combinatoric part of interval exchange transformation can be taken
    independently from its dynamical origin. It has an important link with
    strata of Abelian differential (see :mod:`~sage.combinat.iet.strata`)

    INPUT:

    - ``intervals`` - string, two strings, list, tuples that can be converted to
      two lists

    - ``reduced`` - boolean (default: False) specifies reduction. False means
      labelled permutation and True means reduced permutation.

    - ``flips`` -  iterable (default: None) the letters which correspond to
      flipped intervals.

    - ``alphabet`` - (optional)

    OUTPUT:

    permutation -- the output type depends of the data.

    EXAMPLES::

        sage: from surface_dynamics import *

    Creation of labelled permutations ::

        sage: iet.Permutation('a b c d','d c b a')
        a b c d
        d c b a
        sage: iet.Permutation([[0,1,2,3],[2,1,3,0]])
        0 1 2 3
        2 1 3 0
        sage: iet.Permutation([0, 'A', 'B', 1], ['B', 0, 1, 'A'])
        0 A B 1
        B 0 1 A

    Creation of reduced permutations::

        sage: iet.Permutation('a b c', 'c b a', reduced = True)
        a b c
        c b a
        sage: iet.Permutation([0, 1, 2, 3], [1, 3, 0, 2], reduced=True)
        0 1 2 3
        1 3 0 2
        sage: iet.Permutation([2,1], reduced=True)
        1 2
        2 1

    Managing the alphabet: two labelled permutations with different (ordered)
    alphabet but with the same labels are different::

        sage: p = iet.Permutation('a b','b a', alphabet='ab')
        sage: q = iet.Permutation('a b','b a', alphabet='ba')
        sage: str(p) == str(q)
        True
        sage: p == q
        False
        sage: p.rauzy_move_matrix('top')
        [1 0]
        [1 1]
        sage: q.rauzy_move_matrix('top')
        [1 1]
        [0 1]

    For reduced permutations, the alphabet does not play any role excepted for
    printing the object::

        sage: p = iet.Permutation('a b c','c b a', reduced=True)
        sage: q = iet.Permutation([0,1,2],[2,1,0], reduced=True)
        sage: p == q
        True

    Creation of flipped permutations::

        sage: iet.Permutation('a b c', 'c b a', flips=['a','b'])
        -a -b  c
         c -b -a
        sage: iet.Permutation('a b c', 'c b a', flips='ab', reduced=True)
        -a -b  c
         c -b -a

    TESTS::

        sage: type(iet.Permutation('a b c', 'c b a', reduced=True))
        <class 'surface_dynamics.interval_exchanges.reduced.ReducedPermutationIET'>
        sage: type(iet.Permutation('a b c', 'c b a', reduced=False))
        <class 'surface_dynamics.interval_exchanges.labelled.LabelledPermutationIET'>
        sage: type(iet.Permutation('a b c', 'c b a', reduced=True, flips=['a','b']))
        <class 'surface_dynamics.interval_exchanges.reduced.FlippedReducedPermutationIET'>
        sage: type(iet.Permutation('a b c', 'c b a', reduced=False, flips=['a','b']))
        <class 'surface_dynamics.interval_exchanges.labelled.FlippedLabelledPermutationIET'>

        sage: p = iet.Permutation(('a b c','c b a'))
        sage: iet.Permutation(p) == p
        True
        sage: q = iet.Permutation(p, reduced=True)
        sage: q == p
        False
        sage: q == p.reduced()
        True

        sage: p = iet.Permutation('a', 'a', flips='a', reduced=True)
        sage: iet.Permutation(p) == p
        True

        sage: p = iet.Permutation('a b c','c b a',flips='a')
        sage: iet.Permutation(p) == p
        True
        sage: iet.Permutation(p, reduced=True) == p.reduced()
        True

        sage: p = iet.Permutation('a b c','c b a',reduced=True)
        sage: iet.Permutation(p) == p
        True
    """
    if arg2 is None:
        from .template import Permutation as Permutation_class
        if isinstance(arg1, Permutation_class):
            return Permutation(
                arg1.list(),
                reduced = (arg1._labels is None) if reduced is None else reduced,
                flips = arg1.flips() if flips is None else flips,
                alphabet = arg1.alphabet() if alphabet is None else alphabet)

    if reduced is None:
        reduced = False

    a = _two_lists(arg1, arg2)

    l = a[0] + a[1]
    letters = set(l)

    if flips is not None:
        # make it so that no flip is equivalent to not specifying the flips
        if not flips:
            flips = None
        else:
            for letter in flips:
                if letter not in letters:
                    raise ValueError("flips contains not valid letters")

    for letter in letters:
        if a[0].count(letter) != 1 or a[1].count(letter) != 1:
            raise ValueError("letters must appear once in each interval")

    if reduced:
        if flips is None:
            from .reduced import ReducedPermutationIET as cls
        else :
            from .reduced import FlippedReducedPermutationIET as cls
    else:
        if flips is None:
            from .labelled import LabelledPermutationIET as cls
        else :
            from .labelled import FlippedLabelledPermutationIET as cls

    return cls(a, alphabet=alphabet, reduced=reduced, flips=flips)

def GeneralizedPermutation(arg1, arg2=None, reduced=None, flips=None, alphabet=None):
    r"""
    Returns a permutation of an interval exchange transformation.

    Those permutations are the combinatoric part of linear involutions and were
    introduced by Danthony-Nogueira [DN90]_. The full combinatoric study and
    precise links with strata of quadratic differentials was achieved few years
    later by Boissy-Lanneau [BL08]_.

    INPUT:

    - ``intervals`` - strings, list, tuples

    - ``reduced`` - boolean (defaut: False) specifies reduction. False means
      labelled permutation and True means reduced permutation.

    - ``flips`` -  iterable (default: None) the letters which correspond to
      flipped intervals.

    OUTPUT:

    generalized permutation -- the output type depends on the data.

    EXAMPLES::

        sage: from surface_dynamics import *

    Creation of labelled generalized permutations::

        sage: iet.GeneralizedPermutation('a b b','c c a')
        a b b
        c c a
        sage: iet.GeneralizedPermutation('a a','b b c c')
        a a
        b b c c
        sage: iet.GeneralizedPermutation([[0,1,2,3,1],[4,2,5,3,5,4,0]])
        0 1 2 3 1
        4 2 5 3 5 4 0

    Creation of reduced generalized permutations::

        sage: iet.GeneralizedPermutation('a b b', 'c c a', reduced = True)
        a b b
        c c a
        sage: iet.GeneralizedPermutation('a a b b', 'c c d d', reduced = True)
        a a b b
        c c d d

    Creation of flipped generalized permutations::

        sage: iet.GeneralizedPermutation('a b c a', 'd c d b', flips = ['a','b'])
        -a -b  c -a
         d  c  d -b

    TESTS::

        sage: type(iet.GeneralizedPermutation('a b b', 'c c a', reduced=True))
        <class 'surface_dynamics.interval_exchanges.reduced.ReducedPermutationLI'>
        sage: type(iet.GeneralizedPermutation('a b b', 'c c a', reduced=False))
        <class 'surface_dynamics.interval_exchanges.labelled.LabelledPermutationLI'>
        sage: type(iet.GeneralizedPermutation('a b b', 'c c a', reduced=True, flips=['a','b']))
        <class 'surface_dynamics.interval_exchanges.reduced.FlippedReducedPermutationLI'>
        sage: type(iet.GeneralizedPermutation('a b b', 'c c a', reduced=False, flips=['a','b']))
        <class 'surface_dynamics.interval_exchanges.labelled.FlippedLabelledPermutationLI'>
    """
    if arg2 is None:
        from .template import Permutation as Permutation_class
        if isinstance(arg1, Permutation_class):
            return GeneralizedPermutation(
                arg1.list(),
                reduced = (arg1._labels is None) if reduced is None else reduced,
                flips = arg1.flips() if flips is None else flips,
                alphabet = arg1.alphabet() if alphabet is None else alphabet)

    if reduced is None:
        reduced = False

    a = _two_lists(arg1, arg2)

    l = a[0] + a[1]
    letters = set(l)

    if flips is not None:
        for letter in flips:
            if letter not in letters:
                raise ValueError("flips contains not valid letters")

    for letter in letters:
        if l.count(letter) != 2:
            raise ValueError("letters must appear twice")

    b0 = a[0][:]
    b1 = a[1][:]
    for letter in letters:
        if b0.count(letter) == 1:
            del b0[b0.index(letter)]
            del b1[b1.index(letter)]

    if not b0 and not b1:
        return Permutation(a[0], a[1], reduced=reduced, flips=flips, alphabet=alphabet)

    elif not b0 or not b1:
        raise ValueError("no admissible length")

    if reduced:
        if flips is None:
            from .reduced import ReducedPermutationLI as cls
        else :
            from .reduced import FlippedReducedPermutationLI as cls
    else:
        if flips is None:
            from .labelled import LabelledPermutationLI as cls
        else :
            from .labelled import FlippedLabelledPermutationLI as cls

    return cls(a, alphabet=alphabet, reduced=reduced, flips=flips)

def Permutations_iterator(
    nintervals=None,
    irreducible=True,
    reduced=False,
    alphabet=None):
    r"""
    Returns an iterator over permutations.

    This iterator allows you to iterate over permutations with given
    constraints. If you want to iterate over permutations coming from a given
    stratum you have to use the module :mod:`~sage.combinat.iet.strata` and
    generate Rauzy diagrams from connected components.

    INPUT:

    - ``nintervals`` - non negative integer

    - ``irreducible`` - boolean (default: True)

    - ``reduced`` - boolean (default: False)

    - ``alphabet`` - alphabet (default: None)

    OUTPUT:

    iterator -- an iterator over permutations

    EXAMPLES::

        sage: from surface_dynamics import *

    Generates all reduced permutations with given number of intervals::

        sage: P = iet.Permutations_iterator(nintervals=2,alphabet="ab",reduced=True)
        sage: for p in P: print("%s\n* *" % p)
        a b
        b a
        * *
        sage: P = iet.Permutations_iterator(nintervals=3,alphabet="abc",reduced=True)
        sage: for p in P: print("%s\n* * *" % p)
        a b c
        b c a
        * * *
        a b c
        c a b
        * * *
        a b c
        c b a
        * * *
    """
    from .labelled import LabelledPermutationsIET_iterator
    from .reduced import ReducedPermutationsIET_iterator

    if nintervals is None:
        if alphabet is None:
            raise ValueError("You must specify an alphabet or a length")
        else:
            alphabet = Alphabet(alphabet)
            if alphabet.cardinality() is Infinity:
                raise ValueError("You must sepcify a length with infinite alphabet")
            nintervals = alphabet.cardinality()

    elif alphabet is None:
            alphabet = range(1,nintervals+1)

    if reduced:
        return ReducedPermutationsIET_iterator(
            nintervals,
            irreducible=irreducible,
            alphabet=alphabet)
    else:
        return LabelledPermutationsIET_iterator(
            nintervals,
            irreducible=irreducible,
            alphabet=alphabet)

from sage.misc.decorators import rename_keyword
@rename_keyword(
        lr_inversion='left_right_inversion',
        tb_inversion='top_bottom_inversion')
def RauzyDiagram(arg1, arg2=None, reduced=False, flips=None, alphabet=None,
        right_induction=True, left_induction=False,
        left_right_inversion=False,
        top_bottom_inversion=False,
        symmetric=False):
    r"""
    Return an object coding a Rauzy diagram.

    The Rauzy diagram is an oriented graph with labelled edges. The set of
    vertices corresponds to the permutations obtained by different operations
    (mainly the .rauzy_move() operations that corresponds to an induction of
    interval exchange transformation). The edges correspond to the action of the
    different operations considered.

    It first appeard in the original article of Rauzy [R79]_.

    INPUT:

    - ``intervals`` - lists, or strings, or tuples

    - ``reduced`` - boolean (default: False) to precise reduction

    - ``flips`` - list (default: []) for flipped permutations

    - ``right_induction`` - boolean (default: True) consideration of left
      induction in the diagram

    - ``left_induction`` - boolean (default: False) consideration of right
      induction in the diagram

    - ``left_right_inversion`` - boolean (default: False) consideration of
      inversion

    - ``top_bottom_inversion`` - boolean (default: False) consideration of
      reversion

    - ``symmetric`` - boolean (default: False) consideration of the symmetric
      operation

    OUTPUT:

    Rauzy diagram -- the Rauzy diagram that corresponds to your request

    EXAMPLES::

        sage: from surface_dynamics import *

    Standard Rauzy diagrams::

        sage: iet.RauzyDiagram('a b c d', 'd b c a')
        Rauzy diagram with 12 permutations
        sage: iet.RauzyDiagram('a b c d', 'd b c a', reduced = True)
        Rauzy diagram with 6 permutations

    Extended Rauzy diagrams::

        sage: iet.RauzyDiagram('a b c d', 'd b c a', symmetric=True)
        Rauzy diagram with 144 permutations

    Using Rauzy diagrams and path in Rauzy diagrams::

        sage: r = iet.RauzyDiagram('a b c', 'c b a')
        sage: r
        Rauzy diagram with 3 permutations
        sage: p = iet.Permutation('a b c','c b a')
        sage: p in r
        True
        sage: g0 = r.path(p, 'top', 'bottom','top')
        sage: g1 = r.path(p, 'bottom', 'top', 'bottom')
        sage: g0.is_loop()
        True
        sage: g1.is_loop()
        True
        sage: g0.is_full()
        False
        sage: g1.is_full()
        False
        sage: g = g0 + g1
        sage: g
        Path of length 6 in a Rauzy diagram
        sage: g.is_loop()
        True
        sage: g.is_full()
        True
        sage: m = g.matrix()
        sage: m
        [1 1 1]
        [2 4 1]
        [2 3 2]
        sage: s = g.orbit_substitution()
        sage: print(s)
        a->acbbc, b->acbbcbbc, c->acbc
        sage: s.incidence_matrix() == m
        True

    We can then create the corresponding interval exchange transformation and
    comparing the orbit of `0` to the fixed point of the orbit substitution::

        sage: v = m.eigenvectors_right()[-1][1][0]
        sage: T = iet.IntervalExchangeTransformation(p, v).normalize()
        sage: print(T)
        Interval exchange transformation of [0, 1[ with permutation
        a b c
        c b a
        sage: w1 = []
        sage: x = 0
        sage: for i in range(20):
        ....:  w1.append(T.in_which_interval(x))
        ....:  x = T(x)
        sage: w1 = Word(w1)
        sage: w1
        word: acbbcacbcacbbcbbcacb
        sage: w2 = s.fixed_point('a')
        sage: w2[:20]
        word: acbbcacbcacbbcbbcacb
        sage: w2[:20] == w1
        True
    """
    p = GeneralizedPermutation(
        arg1, arg2,
        reduced = reduced,
        flips = flips,
        alphabet = alphabet)

    return p.rauzy_diagram(
        right_induction = right_induction,
        left_induction = left_induction,
        left_right_inversion = left_right_inversion,
        top_bottom_inversion = top_bottom_inversion,
        symmetric = symmetric)

#TODO
# def GeneralizedPermutation_iterator():
#     print "gpi"

def IntervalExchangeTransformation(permutation=None,lengths=None):
    """
    Constructs an Interval exchange transformation.

    An interval exchange transformation (or iet) is a map from an
    interval to itself. It is defined on the interval except at a finite
    number of points (the singularities) and is a translation on each
    connected component of the complement of the singularities. Moreover it is a
    bijection on its image (or it is injective).

    An interval exchange transformation is encoded by two datas. A permutation
    (that corresponds to the way we echange the intervals) and a vector of
    positive reals (that corresponds to the lengths of the complement of the
    singularities).

    INPUT:

    - ``permutation`` - a permutation

    - ``lengths`` - a list or a dictionnary of lengths

    OUTPUT:

    interval exchange transformation -- an map of an interval

    EXAMPLES::

        sage: from surface_dynamics import *

    Two initialization methods, the first using a iet.Permutation::

        sage: p = iet.Permutation('a b c','c b a')
        sage: t = iet.IntervalExchangeTransformation(p, {'a':1,'b':0.4523,'c':2.8})

    The second is more direct::

        sage: t = iet.IntervalExchangeTransformation(('a b','b a'),{'a':1,'b':4})

    It's also possible to initialize the lengths only with a list::

        sage: t = iet.IntervalExchangeTransformation(('a b c','c b a'),[0.123,0.4,2])

    The two fundamental operations are Rauzy move and normalization::

        sage: t = iet.IntervalExchangeTransformation(('a b c','c b a'),[0.123,0.4,2])
        sage: s = t.rauzy_move()
        sage: s_n = s.normalize(t.length())
        sage: s_n.length() == t.length()
        True

    A not too simple example of a self similar interval exchange transformation::

        sage: p = iet.Permutation('a b c d','d c b a')
        sage: d = p.rauzy_diagram()
        sage: g = d.path(p, 't', 't', 'b', 't', 'b', 'b', 't', 'b')
        sage: m = g.matrix()
        sage: v = m.eigenvectors_right()[-1][1][0]
        sage: t = iet.IntervalExchangeTransformation(p,v)
        sage: s = t.rauzy_move(iterations=8)
        sage: s.normalize() == t.normalize()
        True
    """
    from .iet import IntervalExchangeTransformation as _IET
    from .labelled import LabelledPermutationIET
    from .template import Permutation as Permutation_class

    if isinstance(permutation, Permutation_class) and permutation._flips is not None:
        raise TypeError("interval exchange with flips not yet implemented")
    elif isinstance(permutation, LabelledPermutationIET):
        p = permutation
    elif isinstance(permutation, tuple):
        p = Permutation(*permutation)
    else:
        p = Permutation(permutation)

    if len(lengths) != len(p):
        raise ValueError("bad number of lengths")

    if isinstance(lengths, dict):
        it = lengths.values()
    else:
        it = iter(lengths)
    for x in it:
        try:
            y = float(x)
        except ValueError:
            raise TypeError("unable to convert x (='%s') into a real number" %(str(x)))

        if y < 0:
           raise ValueError("lengths must be non-negative")

    return _IET(p,lengths)

IET = IntervalExchangeTransformation

def IntervalExchangeTransformationFamily(*args):
    r"""
    Return a linear family of interval exchange transformations

    INPUT: either an interval exchange transformation or a pair consisting of a
    permutation and a cone

    EXAMPLES::

        sage: from surface_dynamics import *
        sage: p = iet.Permutation([0,1,2,3,4,5],[5,4,3,2,1,0])
        sage: rays = [[5, 1, 0, 0, 3, 8], [2, 1, 0, 3, 0, 5], [1, 0, 1, 2, 0, 3], [3, 0, 1, 0, 2, 5]]
        sage: F = iet.IETFamily(p, rays)    # optional - pplpy
    """
    try:
        from .iet_family import IETFamily
    except ImportError:
        raise ValueError('flatsurf was compiled without support for iet_family. In order '
                         'to compile flatsurf with iet_family support you need '
                         'to install pplpy with the command:\n    $ sage -pip install pplpy\n'
                         'Once done, you need to reinstall flatsurf')

    if len(args) == 1:
        T = args[0]
        from .iet import IntervalExchangeTransformation
        if not isinstance(T, IntervalExchangeTransformation):
            raise ValueError('not an iet')
        from surface_dynamics.misc.linalg import deformation_cone
        from surface_dynamics.misc.ppl_utils import ppl_convert
        C = ppl_convert(deformation_cone(T.lengths()))
        return IETFamily(T.permutation(), C)

    elif len(args) == 2:
        C = args[1]
        if isinstance(C, (tuple, list)):
            from surface_dynamics.misc.ppl_utils import ppl_cone
            C = ppl_cone(C)
        else:
            from surface_dynamics.misc.ppl_utils import ppl_convert
            C = ppl_convert(C)

        return IETFamily(args[0], C)

IETFamily = IntervalExchangeTransformationFamily



r"""
Exhaustive generation of fat graphs.

This is done following the McKay canonical augmentation. This module
is experimental.
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import absolute_import, print_function
from six.moves import range, map, zip

import numbers

from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ

from sage.stats.basic_stats import mean

from .fat_graph import FatGraph

###########################
# Miscellaneous functions #
###########################

def num_and_weighted_num(it):
    from sage.rings.integer_ring import ZZ
    from sage.rings.rational_field import QQ
    s = QQ.zero()
    n = ZZ.zero()
    for _,aut in it:
        n += ZZ.one()
        if aut is None:
            s += QQ.one()
        else:
            s += QQ((1,aut.group_cardinality()))
    return n,s

def list_extrems(l, n):
    if not n:
        raise ValueError
    vdmin = vdmax = l[0]
    for i in range(1, n):
        if l[i] > vdmax:
            vdmax = l[i]
        if l[i] < vdmin:
            vdmin = l[i]
    return (vdmin, vdmax)


##########################
# Augmentation functions #
##########################

# augment1: trisection
def augment1(cm, aut_grp, g, callback):
    r"""
    Given a unicellular map ``cm`` with a single vertex and automorphism group
    ``aut_grp``, iterate through all its canonical extensions that are
    uniface-univertex maps of greater genus.

    This operation inserts two edes.

    This augmentation function is sufficient to iterate through unicellular map.
    """
    n = cm._n
    fd = cm._fd
    fl = cm._fl
    fp = cm._fp
    i = 0
    if aut_grp is None:
        R = range(n)
    else:
        aut_grp.reset_iterator()
        R = aut_grp

    for i in R:
        j = i
        for sj in range(fd[fl[i]]):
            k = j
            for sk in range(fd[fl[i]] - sj + (i != j)):
                cm.trisect_face(i, j, k)
                test, aaut_grp = cm._is_canonical(n)
                callback('augment1', test, cm, aaut_grp, g-1)
                if test and g>1:
                    augment1(cm, aaut_grp, g - 1, callback)

                cm.remove_face_trisection(n)
                k = fp[k]
            j = fp[j]
        i = fp[i]

# augment2: face split
# (essentially the same as augment3)
def augment2(cm, aut_grp, depth, callback):
    r"""
    Given a map ``cm`` with a single vertex and automorphism group ``aut_grp``
    iterate through all its canonical extensions that are obtained by splitting
    one of its faces (by adding a single edge).

    Because of the chosen canonical labellings, we only need to consider the
    faces with maximal degree and split in such way that the secondly created
    face is still at least as big as the second biggest.
    """
    n = cm._n
    nf = cm._nf
    fp = cm._fp
    fd = cm._fd
    fl = cm._fl

    if aut_grp is None:
        R = range(n)
    else:
        aut_grp.reset_iterator()
        R = aut_grp

    # TODO find edges with the highest face degrees on their sides
    # (the split must remain higher than them)
    fdmax0 = 0
    fdmax1 = 0
    for i in range(nf):
        if fd[i] > fdmax0:
            fdmax1 = fdmax0
            fdmax0 = fd[i]

    if callback is not None:
        parent = cm.to_string()

    for i in R:
        j = i
        if fdmax1 <= 1:
            if fd[fl[i]] < fdmax0 - 1:
                continue
            niter = fd[fl[i]]
        elif fd[fl[i]] != fdmax0 or vd[fl[i]] < 2 * fdmax1 - 2:
            continue
        else:
            for _ in range(fdmax1 - 1):
                j = fp[j]
            niter = fd[fl[i]] - 2*fdmax1 + 3

        for _ in range(niter):
            cm.split_face(i, j)
            test, aaut_grp = cm._is_canonical(n)
            callback('augment2', test, cm, aaut_grp, depth-1)
            if test and depth > 1:
                augment2(cm, aaut_grp, depth-1, callback)
            cm.remove_edge(n)
            j = fp[j]

# augment3: vertex split
def augment3(cm, aut_grp, depth, min_degree, callback):
    r"""
    Given a map ``cm``, its automorphism group ``aut_grp`` and a minimum
    degree ``min_degree``, iterate through all the canonical extensions of
    ``cm`` that are obtained by splitting ``depth`` times a vertices into
    two vertices.

    This augmentation add ``depth`` edges to the fat graph.

    In principle, because of the chosen canonical labellings, we only need to
    consider the vertices with maximal degree.
    """
    n = cm._n
    nv = cm._nv
    vp = cm._vp
    vd = cm._vd
    vl = cm._vl

    if aut_grp is None:
        R = range(n)
    else:
        aut_grp.reset_iterator()
        R = aut_grp

    # TODO: find edges with the highest vertex degrees on their ends
    # the split must remain higher than them
    vdmax0 = 0
    vdmax1 = 0
    for i in range(nv):
        if vd[i] > vdmax0:
            vdmax1 = vdmax0
            vdmax0 = vd[i]
    min_degree_loc = max(min_degree, vdmax1)

    for i in R:
        # vertex degrees are split as d -> (d1 + 1, d2 + 1)
        # so, if min_degree > 1 we can ignore the first/last half edges
        # moreover, by the choosen canonical labelling, the degrees of
        # the split vertices must remain larger than any other

        j = i
        if min_degree_loc == 1:
            if vd[vl[i]] < vdmax0 - 1:
                continue
            niter = vd[vl[i]]
        elif vd[vl[i]] != vdmax0 or vd[vl[i]] < 2 * min_degree_loc - 2:
            continue
        else:
            for _ in range(min_degree_loc - 1):
                j = vp[j]
            niter = vd[vl[i]] - 2*min_degree_loc + 3

        for _ in range(niter):
            cm.split_vertex(i, j)
            assert vd[vl[i]] >= min_degree_loc
            assert vd[vl[j]] >= min_degree_loc
            test, aaut_grp = cm._is_canonical(n)
            callback('augment3', test, cm, aaut_grp, depth-1)
            if test and depth > 1:
                augment3(cm, aaut_grp, depth-1, min_degree, callback)
            cm.contract_edge(n)
            j = vp[j]

# TODO
# def augment4(cm):
#    r"""
#    Plant vertices of degree 1,2.
#    """

########################################
# Callbacks for the various map reduce #
########################################

# callback to count elements (behaves somehow as a 2-tuple)
class CountAndWeightedCount(object):
    def __init__(self):
        self.count = ZZ(0)
        self.weighted_count = QQ(0)
    def __repr__(self):
        return "(%s, %s)" % (self.count, self.weighted_count)
    def __len__(self):
        return 2
    def __getitem__(self, i):
        if not isinstance(i, numbers.Integral):
            raise TypeError
        i = int(i)
        if i == -1 or i == 1:
            return self.weighted_count
        elif i == 0:
            return self.count
        else:
            raise IndexError("index out of range")
    def __eq__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self.count == other.count and self.weighted_count == other.weighted_count
    def __ne__(self, other):
        if type(self) is not type(other):
            raise TypeError
        return self.count != other.count or self.weighted_count != other.weighted_count
    def __call__(self, cm, aut):
        self.count += ZZ(1)
        self.weighted_count += QQ((1, (1 if aut is None else aut.group_cardinality())))

# callback to list elements
class ListCallback(object):
    def __init__(self):
        self._list = []

    def __call__(self, cm, aut):
        self._list.append(cm.copy())

    def list(self):
        return self._list

# TODO: make it work again! This is the most precious piece of information
# to enhance the exhaustive generation...
# Callback for getting a full trace of the execution
grapvhiz_header="""/****************************************************************/
/* Trace execution of fat graphs generation                     */
/*                                                              */
/* root: vp={vp:6} ep={ep:6} fp={fp:6}                          */
/* g = {g:2}                                                       */
/* nf = {nf:2}                                                      */
/* nv = {nv:2}                                                      */
                                                                */
 To compile to a graph in pdf format run                        */
                                                                */
      $ sfpdf -Tpdf -o OUTPUT.pdf INPUT.dot                     */
                                                                */
/****************************************************************/
"""

class FatGraphsTrace(object):
    """
    A class to trace the execution of the fat graphs generation.

    It is mostly used for debugging/profiling/illustration purposes.
    """
    def __init__(self, filename=None, verbosity=0):
        self._verbosity = int(verbosity)
        self._properties = {}
        self._k = 0            # current number of vertices
        self._vdepth = {}      # vertex -> depth
        self._vnum   = {}      # vertex -> apparition in the iteration
        self._edges = {}       # parent -> child
        self._bad_explore = {} # vertex -> number of dead end
        self._vaut = {}        # number of automorphisms

    def __repr__(self):
        return 'FatGraphs trace for {%s}' % (', '.join('%s=%s' % (k,v) for k,v in sorted(self._properties.items())))

    def summary(self, filename=None):
        if filename is None:
            from sys import stdout
            f = stdout
        else:
            f = open(fiename, 'w')

        f.write(repr(self))
        f.write('\n')

        if not self._vdepth:
            if filename is not None:
                f.close()
            return

        count_by_depth = defaultdict(int)
        for v in self._vdepth.values():
            count_by_depth[v] += 1
        max_depth = max(count_by_depth)
        mean_depth = float(sum(k*v for k,v in count_by_depth.items())) / sum(v for v in count_by_depth.values())

        childless_by_depth = defaultdict(int)
        for v,child in self._edges.items():
            if self._vdepth[v] != max_depth and not child:
                childless_by_depth[self._vdepth[v]] += 1

        bad_explore_by_depth = defaultdict(int)
        for v,num in self._bad_explore.items():
            bad_explore_by_depth[self._vdepth[v]] += num

        f.write('depth                   : %d\n' % max_depth)
        f.write('mean depth              : %f\n' % mean_depth)
        f.write('total num fat graphs    : %d\n' % len(self._vdepth))
        f.write('by depth num fat graphs : %s\n' % ' '.join('%d' % count_by_depth[i] for i in range(max_depth + 1)))
        f.write('total bad explore       : %d\n' % sum(bad_explore_by_depth.values()))
        f.write('by depth bad explore    : %s\n' % ' '.join('%d' % bad_explore_by_depth[i] for i in range(max_depth + 1)))
        f.write('total childless         : %d\n' % sum(childless_by_depth.values()))
        f.write('by depth childless      : %s\n' % ' '.join('%d' % childless_by_depth[i] for i in range(max_depth + 1)))

        if filename is not None:
            f.close()

    def __call__(self, cm, aut):
        pass
    def add_vertex(self, s, aut_grp, depth):
        if self._verbosity >= 1:
            print('add_vertex(s={}, depth={})'.format(s, depth))
        if s in self._vdepth:
            raise RuntimeError("already explored vertex!")
        self._vdepth[s] = depth
        self._bad_explore[s] = 0
        self._vnum[s] = self._k
        self._edges[s] = []
        self._k += 1
        self._vaut[s] = 1 if aut_grp is None else aut_grp.group_cardinality()

    def add_edge(self, s0, s1):
        if self._verbosity >= 1:
            print('add_edge(s0={}, s1={})'.format(s0, s1))
        if s0 not in self._edges:
            raise RuntimeError("_edges not properly initialized")
        self._edges[s0].append(s1)

    def root(self, cm, aut_grp):
        if self._k:
            raise RuntimeError("trying to set root in a non-empty trace")
        s = cm.to_string()
        if self._verbosity >= 1:
            print('root(cm={})'.format(s))
        self.add_vertex(s, aut_grp, 0)

    def canonical_edge(self, parent, cm, aut_grp, caller):
        if not isinstance(parent, str):
            raise RuntimeError

        s = cm.to_string()
        if self._verbosity >= 1:
            print('canonical_edge(parent={}, cm={}, caller={})'.format(parent, s, caller))

        if parent not in self._edges:
            raise RuntimeError("_edges not properly initialized at %s" % parent)
        if s in self._edges:
            raise RuntimeError("s already in _edges")

        self.add_vertex(s, aut_grp, self._vdepth[parent] + 1)
        self.add_edge(parent, s)

    def non_canonical_edge(self, parent, cm, caller):
        if not isinstance(parent, str):
            raise RuntimeError
        if self._verbosity >= 1:
            print('non_canonical_edge(parent={}, caller={})'.format(parent, caller))
        if parent not in self._edges:
            raise RuntimeError("_edges not properly initialized at %s" % parent)
        self._bad_explore[parent] += 1

    def grapvhiz_tree(self, filename=None):
        if filename is None:
            from sys import stdout
            output = stdout
        else:
            output = open(filename, 'w')

        if filename is not None:
            output.close()

        f = open(filename, 'w')
        f.write(header.format(vp=vp, ep=ep, fp=fp, g=g, nf=nf, nv=nv))

        f.write('digraph Tree {\n')
        f.write('   rankdir = TB;\n')

        col1 = "#FF0000"
        col2 = "#00FF00"
        col3 = "#0000FF"

        a0 = cm0.automorphism_group()
        s0 = cm0.to_string()
        f.write("""   %s [label="%s"];\n""" % (s0, 0))
        for cm1, a1 in augment1(cm0, a0, g, False):
            s1 = cm1.to_string()
            if s0 != s1:
                f.write("""   %s [label="%s"];\n""" % (s1, 0))
                f.write("""   %s -> %s [color="%s"];\n""" %(s0, s1, col1))
            for cm2, a2 in augment2(cm1, a1, nnf, intermediate):
                s2 = cm2.to_string()
                if s2 != s1:
                    f.write("""   %s [label="%s"];\n""" % (s2, 0))
                    f.write("""   %s -> %s [color="%s"];\n""" %(s1, s2, col2))
                for cm3, a3 in augment3(cm2, a2, nnv, vertex_min_degree, intermediate):
                    s3 = cm3.to_string()
                    if s3 != s2:
                        f.write("""   %s [label="%s"];\n""" % (s3, 0))
                        f.write("""   %s -> %s [color="%s"];\n""" %(s2, s3, col3))
                    yield cm3, a3

#################
# Main iterator #
#################

class StackCallback(object):
    def __init__(self, cm, aut, gmin, gdepth, nfmin, nfdepth, nvmin, nvdepth, min_degree, callback, filter):
        self._gmin = gmin
        self._gdepth = gdepth
        self._nfmin = nfmin 
        self._nfdepth = nfdepth
        self._nvmin = nvmin
        self._nvdepth = nvdepth
        self._cm = cm
        self._aut = aut
        self._callback = callback
        self._min_degree = min_degree
        self._filter = filter
        # print("StackCallback(gmin={}, gdepth={}, nfmin={}, nfdepth={}, nvmin={}, nvdepth={}".format(
        #    self._gmin, self._gdepth, self._nfmin, self._nfdepth, self._nvmin, self._nvdepth))

    def __call__(self, caller, test, cm, aut, depth):
        if test:
            if caller == 'augment1':
                # we know the map has one vertex and one edge and hence 4g = n
                if cm._n >= 4 * self._gmin:
                    if cm._nv >= self._nvmin and cm._nf >= self._nfmin and \
                       (self._filter is None or self._filter(cm,aut)):
                        self._callback(cm, aut)

                    # more faces?
                    if self._nfdepth:
                        augment2(cm, aut, self._nfdepth, self)
                    # more vertices?
                    elif self._nvdepth:
                        augment3(cm, aut, self._nvdepth, self._min_degree, self)

            elif caller == 'augment2':
                if cm._nf >= self._nfmin:
                    if cm._nv >= self._nvmin and \
                       (self._filter is None or self._filter(cm, aut)):
                        self._callback(cm, aut)

                    # more vertices?
                    if self._nvdepth:
                        augment3(cm, aut, self._nvdepth, self._min_degree, self)

            elif caller == 'augment3':
                if cm._nv >= self._nvmin and \
                   (self._filter is None or self._filter(cm, aut)):
                    self._callback(cm, aut)

    def run(self):
        if self._gdepth:
            augment1(self._cm, self._aut, self._gdepth, self)
        elif self._nfdepth:
            augment2(self._cm, self._aut, self._nfdepth, self)
        elif self._nvdepth:
            augment3(self._cm, self._aut, self._nvdepth, self._min_degree, self)
        elif self._filter is None or self._filter(cm, aut):
            self._callback(self._cm, self._aut)

##############
# Main class #
##############

class FatGraphs_g_nf_nv(object):
    r"""
    Isomorphism classes of fat graphs with given genus, number of faces and
    number of vertices.

    EXAMPLES::

        sage: from surface_dynamics.topology.fat_graph_exhaustive_generation import FatGraphs_g_nf_nv

    Trees and their dual (maps with single vertex) in genus zero are counted by
    Catalan numbers::

        sage: for n in range(2, 10):
        ....:     ntrees1 = 2 * (n-1) * FatGraphs_g_nf_nv(0, n, 1).weighted_cardinality()
        ....:     ntrees2 = 2 * (n-1) * FatGraphs_g_nf_nv(0, 1, n).weighted_cardinality()
        ....:     assert catalan_number(n-1) == ntrees1 == ntrees2, n

    Genus zero with same number of vertices and faces::
        
        sage: FatGraphs_g_nf_nv(0, 2, 2).cardinality_and_weighted_cardinality()
        (2, 5/4)
        sage: FatGraphs_g_nf_nv(0, 3, 3).cardinality_and_weighted_cardinality()
        (23, 41/2)
        sage: FatGraphs_g_nf_nv(0, 4, 4).cardinality_and_weighted_cardinality()
        (761, 8885/12)

    Duality checks

        sage: for g,nf,nv in [(0,2,3), (0,2,4), (0,3,4),
        ....:                 (1,1,2), (1,1,3), (1,1,4), (1,1,5), (1,2,3), (1,2,4), (1,3,4),
        ....:                 (2,1,2), (2,1,3)]:
        ....:     n1 = FatGraphs_g_nf_nv(g, nf, nv).cardinality_and_weighted_cardinality()
        ....:     n2 = FatGraphs_g_nf_nv(g, nv, nf).cardinality_and_weighted_cardinality()
        ....:     assert n1 == n2
        ....:     print(g, nf, nv, n1)
        0 2 3 (5, 11/3)
        0 2 4 (14, 93/8)
        0 3 4 (108, 103)
        1 1 2 (3, 5/3)
        1 1 3 (11, 35/4)
        1 1 4 (46, 42)
        1 1 5 (204, 385/2)
        1 2 3 (180, 172)
        1 2 4 (1198, 14065/12)
        1 3 4 (18396, 18294)
        2 1 2 (53, 483/10)
        2 1 3 (553, 539)

    Unicellular map with one vertex in genus 3::

        sage: FatGraphs_g_nf_nv(3, 1, 1).cardinality_and_weighted_cardinality()
        (131, 495/4)

    Minimum vertex degree bounds::

        sage: for k in range(2,5):
        ....:    F = FatGraphs_g_nf_nv(1, 2, 2, vertex_min_degree=1)
        ....:    c1 = F.cardinality_and_weighted_cardinality(lambda cm,_: cm.vertex_degree_min() >= k)
        ....:    G = FatGraphs_g_nf_nv(1, 2, 2, vertex_min_degree=k)
        ....:    c2 = G.cardinality_and_weighted_cardinality()
        ....:    assert c1 == c2
        ....:    print(c1)
        (14, 87/8)
        (8, 47/8)
        (4, 15/8)

        sage: for k in range(2,6):
        ....:    F = FatGraphs_g_nf_nv(0, 5, 2, vertex_min_degree=1)
        ....:    c1 = F.cardinality_and_weighted_cardinality(lambda cm,_: cm.vertex_degree_min() >= k)
        ....:    G = FatGraphs_g_nf_nv(0, 5, 2, vertex_min_degree=k)
        ....:    c2 = G.cardinality_and_weighted_cardinality()
        ....:    assert c1 == c2
        ....:    print(c1)
        (28, 123/5)
        (21, 88/5)
        (13, 48/5)
        (7, 18/5)

    Using ranges for vertex and face numbers::

        sage: F = FatGraphs_g_nf_nv(1, nv_min=2, nv_max=4, nf_min=2, nf_max=4)
        sage: def check(cm,aut):
        ....:    if cm.num_vertices() < 2 or \
        ....:       cm.num_vertices() >= 4 or \
        ....:       cm.num_faces() < 2 or \
        ....:       cm.num_faces() >= 4:
        ....:           raise ValueError(str(cm))
        sage: F.map_reduce(check)
        sage: for nf in [2,3]:
        ....:     for nv in [2,3]:
        ....:         c1 = F.cardinality_and_weighted_cardinality(lambda cm,_: cm.num_vertices() == nv and cm.num_faces() == nf)
        ....:         c2 = FatGraphs_g_nf_nv(1, nf, nv).cardinality_and_weighted_cardinality()
        ....:         assert c1 == c2
        ....:         print(nf, nv, c1)
        2 2 (24, 167/8)
        2 3 (180, 172)
        3 2 (180, 172)
        3 3 (2048, 6041/3)
    """
    def __init__(self, g=None, nf=None, nv=None, vertex_min_degree=1, g_min=None, g_max=None, nf_min=None, nf_max=None, nv_min=None, nv_max=None):
        r"""
       INPUT:

        - ``g``, ``g_min``, ``g_max`` - the genus

        - ``nf``, ``nf_min``, ``nf_max`` - number of faces

        - ``nv``, ``nv_min``, ``nv_max`` - number of vertices

        - ``vertex_min_degree`` - minimal number of vertices (default to ``1``)

        - ``intermediate`` - if set to ``True`` then return all graphs with genus = g,
          number of faces <= nf and number of vertices <= nv that satisfy the
          ``vertex_min_degree`` constraint
        """
        self._gmin, self._gmax = self._get_interval(g, g_min, g_max, 0, 'g')
        if self._gmax != self._gmin + 1:
            raise ValueError("not implemented for genus in an interval")
        self._fmin, self._fmax = self._get_interval(nf, nf_min, nf_max, 1, 'nf')
        self._vmin, self._vmax = self._get_interval(nv, nv_min, nv_max, 1, 'nv')
        self._vertex_min_degree = ZZ(vertex_min_degree)

    def _get_interval(self, v, vmin, vmax, low_bnd, name):
        if v is not None:
            if not isinstance(v, numbers.Integral):
                raise TypeError("%s must be an integral" % name)
            v = ZZ(v)
            if v < low_bnd:
                raise ValueError("%s must be >= %d" % (name, low_bnd))
            return (v, v+1)
        if vmax is None:
            raise ValueError("at least %s or %s_max must be set" % (name, name))
        if not isinstance(vmax, numbers.Integral):
            raise ValueError("%s_max must be integral" % name)
        vmax = ZZ(vmax)
        if vmin is None:
            vmin = low_bnd
        elif not isinstance(vmin, numbers.Integral):
            raise TypeError("%s_min must be integral" % name)
        vmin = ZZ(vmin)
        if vmin < low_bnd:
            raise ValueError("%s_min must be >= %s" % (name, low_bnd))
        return vmin, vmax

    def __repr__(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph_exhaustive_generation import FatGraphs_g_nf_nv
            sage: FatGraphs_g_nf_nv(0, 5, 1, vertex_min_degree=3)
            Fat graphs of genus 0, with 5 faces and 1 vertices
            sage: FatGraphs_g_nf_nv(0, nf_min=2, nf_max=5, nv_min=1, nv_max=5, vertex_min_degree=3)
            Fat graphs of genus 0, with between 2 and 5 faces and between 1 and 5 vertices
        """
        if self._gmax == self._gmin + 1:
            genus = '%d' % self._gmin
        else:
            genus = 'between %d and %d' % (self._gmin, self._gmax)

        if self._fmax == self._fmin + 1:
            faces = '%d' % self._fmin
        else:
            faces = 'between %d and %d' % (self._fmin, self._fmax)

        if self._vmax == self._vmin + 1:
            vertices = '%d' % self._vmin
        else:
            vertices = 'between %d and %d' % (self._vmin, self._vmax)

        return "Fat graphs of genus %s, with %s faces and %s vertices" % (genus, faces, vertices)

    def map_reduce(self, callback, filter=None):
        r"""
        EXAMPLES::

            sage: from __future__ import print_function
            sage: from surface_dynamics.topology.fat_graph_exhaustive_generation import FatGraphs_g_nf_nv
            sage: FatGraphs_g_nf_nv(1, 2, 2).map_reduce(lambda x,y: print(x))
            FatGraph('(0,6,5,4,1,2,3)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,3,4,6,7)(5)')
            FatGraph('(0,6,1,2,3)(4,7,5)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,3,6,4,7)(5)')
            FatGraph('(0,5,4,1,6,2,3)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,6,7,1,2,3,4)(5)')
            FatGraph('(0,7,2,3)(1,6,5,4)', '(0,2)(1,3)(4,5)(6,7)', '(0,7,1,2,3,4,6)(5)')
            FatGraph('(0,5,4,1,2,6,3)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,6,7,2,3,4)(5)')
            FatGraph('(0,7,3)(1,2,6,5,4)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,7,2,3,4,6)(5)')
            FatGraph('(0,5,4,1,2,3,6)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,6,7,3,4)(5)')
            FatGraph('(0,7)(1,2,3,6,5,4)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,7,3,4,6)(5)')
            FatGraph('(0,5,4,6,1,2,3)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,3,6,7,4)(5)')
            FatGraph('(0,5,4,6,2,3)(1,7)', '(0,2)(1,3)(4,5)(6,7)', '(0,6,1,2,3,7,4)(5)')
            FatGraph('(0,5,6,4,1,2,3)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,3,4)(5,6,7)')
            FatGraph('(0,5,6,1,2,3)(4,7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,3,6,4)(5,7)')
            FatGraph('(0,5,6,2,3)(1,7,4)', '(0,2)(1,3)(4,5)(6,7)', '(0,6,1,2,3,4)(5,7)')
            FatGraph('(0,5,6,3)(1,2,7,4)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,6,2,3,4)(5,7)')
            FatGraph('(0,6,5,1,2,4,3)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,4,6,7)(2,3,5)')
            FatGraph('(0,6,1,2,4,3)(5,7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,4,7)(2,3,6,5)')
            FatGraph('(0,6,4,3)(1,2,7,5)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,4,7)(2,3,5,6)')
            FatGraph('(0,6,5,1,2,3,4)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,4,6,7)(3,5)')
            FatGraph('(0,5,1,6,2,3,4)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,6,7,1,2,4)(3,5)')
            FatGraph('(0,5,1,6,3,4)(2,7)', '(0,2)(1,3)(4,5)(6,7)', '(0,7,1,6,2,4)(3,5)')
            FatGraph('(0,5,1,2,3,6,4)(7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,4)(3,5,6,7)')
            FatGraph('(0,5,1,2,3,6)(4,7)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,6,4)(3,5,7)')
            FatGraph('(0,7,4)(1,2,3,6,5)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,4,6)(3,5,7)')
            FatGraph('(0,5,7,4)(1,2,3,6)', '(0,2)(1,3)(4,5)(6,7)', '(0,1,2,4)(3,6,5,7)')
        """
        g = self._gmin
        fmin = self._fmin
        fmax = self._fmax
        vmin = self._vmin
        vmax = self._vmax
        if g == 0:
            if fmin == 1 and vmin == 1:
                raise NotImplementedError
            elif fmin > 1:
                # start with face splitting of the bicellular map (g = 0, nv = 1, nf = 2)
                cm0 = FatGraph('(0,1)', '(0,1)', '(0)(1)')
                gshift = 0
                vshift = - 1
                fshift = - 2
            elif vmin > 1:
                # start with vertex splitting of the unicellular map (g = 0, nv = 2, nf = 1)
                cm0 = FatGraph('(0)(1)', '(0,1)', '(0,1)')
                gshift = 0
                vshift = - 2
                fshift = - 1
            else:
                raise RuntimeError("this should not happen")
        else:
            # start with trisection of the trivial map (g = 1, nv = 1, nf = 1)
            cm0 = FatGraph.from_unicellular_word([0,1,0,1])
            gshift = - 1
            vshift = - 1
            fshift = - 1
        cm0._realloc(4 * g + 2 * (fmax + vmax - 2))

        a0 = cm0.automorphism_group()
        StackCallback(cm0, a0,
                g,                  # gmin
                g + gshift,         # depth for trisections
                fmin,               # nfmin
                fmax + fshift - 1,  # depth for face splitting
                vmin,               # nvmin
                vmax + vshift - 1,  # depth for vertex splitting
                self._vertex_min_degree,
                callback,
                filter).run()

    def cardinality_and_weighted_cardinality(self, filter=None):
        N = CountAndWeightedCount()
        self.map_reduce(N, filter)
        return tuple(N)

    def weighted_cardinality(self, filter=None):
        N = CountAndWeightedCount()
        self.map_reduce(N, filter)
        return N[1]

    def list(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics.topology.fat_graph_exhaustive_generation import FatGraphs_g_nf_nv
            sage: L21 = FatGraphs_g_nf_nv(0, 2, 1).list()
            sage: L21[0].num_faces()
            2
            sage: L21[0].num_vertices()
            1


            sage: L12 = FatGraphs_g_nf_nv(0, 1, 2).list()
            sage: L12[0].num_faces()
            1
            sage: L12[0].num_vertices()
            2
        """
        L = ListCallback()
        self.map_reduce(L)
        return L.list()




#*****************************************************************************
#       Copyright (C) 2020 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************
from sage.misc.cachefunc import cached_function
from sage.rings.all import ZZ, QQ
from sage.arith.misc import bernoulli, factorial
ZZ_0 = ZZ.zero()
ZZ_1 = ZZ.one()
ZZ_2 = ZZ(2)

@cached_function
def zeta_no_pi(k):
    r"""
    zeta(2m+2) = (-1)^m B_{2m+2} (2pi)^(2m+2) / (2 (2m+2)!)

    EXAMPLES::

        sage: from surface_dynamics.topological_recursion.no_pi import zeta_no_pi
        sage: zeta_no_pi(2)
        1/6
        sage: zeta_no_pi(4)
        1/90
        sage: zeta_no_pi(6)
        1/945

        sage: zeta(2)
        1/6*pi^2
        sage: zeta(4)
        1/90*pi^4
        sage: zeta(6)
        1/945*pi^6
    """
    k = ZZ(k)
    if k % 2 == 1:
        raise ValueError
    m = (k - 2) // 2
    return ZZ(-1)**m * bernoulli(k) * ZZ_2**(k-1) / k.factorial()




# distutils: language=c++
# distutils: extra_compile_args=-std=c++11
# distutils: libraries=ppl gmp
r"""
Linear families of interval exchange transformations
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

##########################
# C declarations
##########################

cdef extern from "gmp.h":
    ctypedef unsigned long mp_limb_t
    ctypedef long mp_size_t

    ctypedef struct __mpz_struct:
        pass
    ctypedef __mpz_struct *mpz_srcptr
    ctypedef __mpz_struct mpz_t[1]

    void mpz_init (mpz_t integer) nogil
    void mpz_clear (mpz_t integer) nogil

    void mpz_set (mpz_t rop, mpz_t op)
    void mpz_set_ui (mpz_t rop, unsigned long int op) nogil
    void mpz_set_si (mpz_t rop, signed long int op) nogil
    void mpz_add (mpz_t rop, mpz_t op1, mpz_t op2) nogil
    int mpz_cmp (mpz_t op1, mpz_t op2) nogil
    int mpz_sgn (mpz_t op)

    size_t mpz_size (mpz_t op)
    mp_limb_t mpz_getlimbn (mpz_t op, mp_size_t n)

########################
# Cython imports
########################

from libc.stdlib cimport malloc, calloc, realloc, free, qsort
from libc.string cimport memcpy

from cpython.object cimport Py_EQ, Py_NE
from cython.operator cimport dereference as deref

from cysignals.memory cimport sig_malloc, sig_calloc, sig_realloc, sig_free, check_malloc, check_calloc, check_realloc

from ppl.ppl_decl cimport PPL_Variable, PPL_Generator, PPL_Generator_System, PPL_gs_iterator
from ppl.linear_algebra cimport Variable
from ppl.generator cimport Generator_System
from ppl.polyhedron cimport C_Polyhedron

from sage.ext.stdsage cimport PY_NEW
from sage.rings.integer cimport Integer

#from gmpy2 cimport import_gmpy2, MPZ_Object, MPZ, GMPy_MPZ_New
#import_gmpy2()


########################
# Begining of the code
########################

cdef int compare(const void * _left, const void * _right) nogil:
    cdef mpz_t * left = (<mpz_t **> _left)[0]
    cdef mpz_t * right = (<mpz_t **> _right)[0]
    cdef int c = mpz_cmp(left[0], right[0])
    while c == 0:
        left += 1
        right += 1
        c = mpz_cmp(left[0], right[0])
    return c

cdef inline int mpz_vec_any_smaller(mpz_t * x, mpz_t * y, size_t length):
    r"""
    Check whether any x[i] < y[i]
    """
    cdef size_t k
    for k in range(length):
        if mpz_cmp(x[k], y[k]) < 0:
            return 1
    return 0

cdef inline int mpz_vec_equal(mpz_t * x, mpz_t * y, size_t length):
    cdef size_t k
    for k in range(length):
        if mpz_cmp(x[k], y[k]): return 0
    return 1

cdef Py_hash_t mpz_pythonhash(mpz_srcptr z):
    """
    Hash an ``mpz``, where the hash value is the same as the hash value
    of the corresponding Python ``long``, except that we do not replace
    -1 by -2 (the Cython wrapper for ``__hash__`` does that).
    """
    # Add all limbs, adding 1 for every carry
    cdef mp_limb_t h1 = 0
    cdef mp_limb_t h0
    cdef size_t i, n
    n = mpz_size(z)
    for i in range(n):
        h0 = h1
        h1 += mpz_getlimbn(z, i)
        # Add 1 on overflow
        if h1 < h0: h1 += 1

    cdef Py_hash_t h = h1
    if mpz_sgn(z) < 0:
        return -h
    return h

cdef inline void rauzy_top(int * perm, size_t dim):
    cdef int a = perm[dim - 1]
    cdef size_t i, j
    for i in range(dim, 2*dim):
        if perm[i] == a:
            break
    a = perm[2 * dim - 1]
    for j in range(2 * dim - 1, i + 1, -1):
        perm[j] = perm[j - 1]
    perm[i + 1] = a

cdef inline void rauzy_bot(int * perm, size_t dim):
    cdef int a = perm[2*dim - 1]
    cdef size_t i, j
    for i in range(dim):
        if perm[i] == a:
            break
    a = perm[dim - 1]
    for j in range(dim - 1, i + 1, -1):
        perm[j] = perm[j - 1]
    perm[i + 1] = a

cdef class IETFamily(object):
    r"""
    A linear family of interval exchange transformations

    This class consists of two attributes:

    - a permutation (must be labeled)

    - a cone of lengths (a subcone of the positive cone)

    EXAMPLES::

        sage: from surface_dynamics import *

        sage: p = iet.Permutation('a b c d', 'd c b a')
        sage: F = iet.IETFamily(p, (ZZ**4).basis())    # optional - pplpy
        sage: F  # optional - pplpy
        top 0 1 2 3
        bot 3 2 1 0
        0 0 0 1
        0 0 1 0
        0 1 0 0
        1 0 0 0

        sage: iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy
        top 0 1 2 3
        bot 3 2 1 0
        0 1 1 1
        2 3 0 0

        sage: iet.IETFamily(p, [(1,0,0,0), (1,-1,1,1)]) # optional - pplpy
        Traceback (most recent call last):
        ...
        ValueError: C must be a subcone of the non-negative cone
        sage: iet.IETFamily(p, Polyhedron(vertices=[(0,0,0,0), (1,0,0,0),(0,1,0,0),(1,1,1,1)])) # optional - pplpy
        Traceback (most recent call last):
        ...
        ValueError: should have only zero as vertices
    """
    cdef int * perm         # top perm
    cdef size_t dim         # dimension (= number of letters)
    cdef mpz_t * entries    # ray coefficients
    cdef mpz_t ** rows      # ray pointers to entries (begining of vectors)
    cdef size_t alloc       # ray allocation size
    cdef size_t length      # number of rays
    cdef C_Polyhedron poly  # the polyhedron (that we should get rid of)

    def __cinit__(self):
        self.perm = NULL
        self.entries = NULL
        self.rows = NULL
        self.alloc = 0
        self.dim = 0
        self.length = 0

    def __dealloc__(self):
        cdef size_t i
        for i in range(self.dim * self.alloc):
            mpz_clear(self.entries[i])
        free(self.perm)
        free(self.entries)
        free(self.rows)

    def __init__(self, p, C_Polyhedron C):
        # convert and check input
        from surface_dynamics.interval_exchanges.labelled import LabelledPermutationIET
        if not isinstance(p, LabelledPermutationIET):
            raise ValueError('p must be a labelled permutation')

        from surface_dynamics.misc.ppl_utils import ppl_check_non_negative_cone
        ppl_check_non_negative_cone(C)
        if C.space_dimension() != len(p):
            raise ValueError('dimension mismatch')

        # fill C data from input
        self.dim = len(p)
        self.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
        cdef size_t i
        cdef int j
        for i,j in enumerate(p._labels[0]):
            self.perm[i] = j
        for i,j in enumerate(p._labels[1]):
            self.perm[self.dim + i] = j

        self.poly = <C_Polyhedron> C
        self.fill_rays(self.poly.thisptr.minimized_generators())

    def ray_coefficient(self, size_t nray, size_t i):
        r"""
        Return the ``i``-th coefficient of the ``nray``-th ray

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: F = iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy
            sage: F.ray_coefficient(0, 2) # optional - pplpy
            1
            sage: F.ray_coefficient(1, 0) # optional - pplpy
            2
            sage: F.ray_coefficient(3, 0) # optional - pplpy
            Traceback (most recent call last):
            ...
            IndexError: row index out of range
            sage: F.ray_coefficient(1, 5) # optional - pplpy
            Traceback (most recent call last):
            ...
            IndexError: column index out of range
        """
        if nray >= self.length:
            raise IndexError('row index out of range')
        if i >= self.dim:
            raise IndexError('column index out of range')
        cdef Integer z= PY_NEW(Integer)
        mpz_set(z.value, self.rows[nray][i])
        return z

    def rays(self):
        r"""
        Return the rays as vectors

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: F = iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy
            sage: F.rays() # optional - pplpy
            [(0, 1, 1, 1), (2, 3, 0, 0)]
        """
        from sage.modules.free_module import FreeModule
        from sage.rings.rational_field import QQ
        F = FreeModule(QQ, self.dim)
        return [F([self.ray_coefficient(i, j) for j in range(self.dim)]) for i in range(self.length)]

    def __repr__(self):
        r"""
        TESTS::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: F = iet.IETFamily(p, [(2,3,0,0), (0,1,1,1)]) # optional - pplpy
            sage: repr(F)  # indirect doctest # optional - pplpy
            'top 0 1 2 3\nbot 3 2 1 0\n0 1 1 1\n2 3 0 0'
        """
        cdef size_t i, j
        s = []
        s.append('top ' + ' '.join(str(self.perm[i]) for i in range(self.dim)))
        s.append('bot ' + ' '.join(str(self.perm[self.dim+i]) for i in range(self.dim)))
        for i in range(self.length):
            s.append(' '.join(str(self.ray_coefficient(i, j)) for j in range(self.dim)))
        return '\n'.join(s)

    def _debug_info(self):
        import sys
        sys.stdout.write('dim      %lu\n' % self.dim)
        sys.stdout.write('length   %lu\n' % self.length)
        sys.stdout.write('alloc    %lu\n' % self.alloc)
        sys.stdout.flush()
        sys.stdout.write('entries  %lu\n' % <size_t> self.entries)
        sys.stdout.flush()
        sys.stdout.write('rows     %lu\n' % <size_t> self.rows)
        sys.stdout.flush()
        cdef size_t i
        for i in range(self.alloc):
            sys.stdout.write('rows[%d]  %lu\n' %(i, <size_t> (self.rows[i] - self.entries)))
            sys.stdout.flush()


    cdef void fit_length(self, size_t length):
        r"""
        Manage allocation

        After this function is called, it is ensured that there is enough space
        for ``length`` rays.
        """
        cdef size_t i
        cdef mpz_t * old_entries
        cdef ptrdiff_t offset
        if length > self.alloc:
            length = max(length, 2 * self.alloc)

            old_entries = self.entries
            self.entries = <mpz_t *> realloc(self.entries, self.dim * length * sizeof(mpz_t))
            offset = self.entries - old_entries

            for i in range(self.dim * self.alloc, self.dim * length):
                mpz_init(self.entries[i])

            self.rows = <mpz_t **> realloc(self.rows, length * sizeof(mpz_t *))
            if offset:
                for i in range(self.alloc):
                    self.rows[i] += offset
            for i in range(self.alloc, length):
                self.rows[i] = self.entries + (i * self.dim)
            self.alloc = length

    cdef void fill_rays(self, const PPL_Generator_System& gs):
        r"""
        Set the attributes related to the cone
        """
        cdef PPL_gs_iterator * gsi_ptr = new PPL_gs_iterator(gs.begin())
        cdef size_t j
        while gsi_ptr[0] != gs.end():
            if deref(gsi_ptr[0]).is_ray():
                self.fit_length(self.length + 1)
                for j in range(self.dim):
                    mpz_set(self.rows[self.length][j],
                            deref(gsi_ptr[0]).coefficient(PPL_Variable(j)).get_mpz_t())
                self.length += 1
            gsi_ptr[0].inc(1)
        del gsi_ptr
        qsort(self.rows, self.length, sizeof(mpz_t *), & compare)

    def permutation(self):
        r"""
        Return the permutation as a list of integers
        """
        cdef size_t i
        return [[self.perm[i] for i in range(self.dim)], [self.perm[self.dim+i] for i in range(self.dim)]]

    def __hash__(self):
        r"""
        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: F = iet.IETFamily(p, Polyhedron(rays=(ZZ**4).basis())) # optional - pplpy
            sage: hash(F) # optional - pplpy
            4696861030007372737
        """
        cdef Py_hash_t x, y, mult
        cdef size_t i, j

        mult = 3
        x = 0x345678
        for i in range(self.dim):
            x = (x ^ (self.perm[i])) * mult
            mult += 1345
            x += 35

        for i in range(self.length):
            for j in range(self.dim):
                y = mpz_pythonhash(self.rows[i][j])
                x = (x ^ y) * mult
                mult += 82520
            x += 97531

        return x

    def __richcmp__(IETFamily self, IETFamily other, int op):
        r"""
        TESTS::

            sage: from surface_dynamics import *

            sage: p1 = iet.Permutation('a b c d', 'd c b a')
            sage: p2 = iet.Permutation('a b c d', 'd a b c')
            sage: rays1 = [[0,1,0,1],[1,0,1,0],[1,1,0,0],[0,0,1,1]]
            sage: rays2 = [[1,0,0,0],[1,0,1,0],[1,1,0,0],[0,0,1,1]]
            sage: rays3 = [[1,0,1,0],[1,1,0,0],[0,0,1,1]]

            sage: iet.IETFamily(p1, rays1) == iet.IETFamily(p1, rays1) # optional - pplpy
            True
            sage: iet.IETFamily(p1, rays1) != iet.IETFamily(p1, rays1) # optional - pplpy
            False

            sage: iet.IETFamily(p1, rays1) == iet.IETFamily(p2, rays1) # optional - pplpy
            False
            sage: iet.IETFamily(p1, rays1) != iet.IETFamily(p2, rays1) # optional - pplpy
            True

            sage: iet.IETFamily(p1, rays1) == iet.IETFamily(p2, rays1) # optional - pplpy
            False
            sage: iet.IETFamily(p1, rays1) != iet.IETFamily(p2, rays1) # optional - pplpy
            True

            sage: iet.IETFamily(p1, rays1) == iet.IETFamily(p1, rays3) # optional - pplpy
            False
            sage: iet.IETFamily(p1, rays1) != iet.IETFamily(p1, rays3) # optional - pplpy
            True

            sage: iet.IETFamily(p1, rays1) == iet.IETFamily(p2, rays3) # optional - pplpy
            False
            sage: iet.IETFamily(p1, rays1) != iet.IETFamily(p2, rays3) # optional - pplpy
            True
        """
        if op != Py_EQ and op != Py_NE:
            raise TypeError('only equality and difference tests available')
        if self.dim != other.dim or self.length != other.length:
            return op == Py_NE

        cdef size_t i,j
        for i in range(2 * self.dim):
            if self.perm[i] != other.perm[i]:
                return op == Py_NE
        for i in range(self.length):
            for j in range(self.dim):
                if mpz_cmp(self.rows[i][j], other.rows[i][j]):
                    return op == Py_NE

        return op == Py_EQ

    # this is completely useless and checks many times the same thing
    # we should just look at left/right Rauzy induction
    def has_connection(self):
        r"""
        Check whether this family has a connection.

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation('A B C D E', 'E D C B A')
            sage: C0 = Polyhedron(rays=(ZZ**5).basis())
            sage: iet.IETFamily(p, C0).has_connection() # optional - pplpy
            False

            sage: C = Polyhedron(rays=[(1,2,3,2,3),(1,1,0,0,2)])
            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy
            True

            sage: C = Polyhedron(rays=[(1,1,1,1,4),(2,0,1,2,5)])
            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy
            True

            sage: C = Polyhedron(rays=[(1,0,0,1,0),(1,0,0,0,1),(0,1,0,1,0),(0,1,0,0,1)])
            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy
            True

            sage: C = Polyhedron(rays=[(1,2,2,2,2)])
            sage: iet.IETFamily(p, C).has_connection() # optional - pplpy
            False
        """
        cdef bint ans

        cdef mpz_t * x = <mpz_t *> check_malloc(2 * self.length * sizeof(mpz_t))
        cdef mpz_t * y = x + self.length
        cdef size_t i, j, k
        for i in range(self.length):
            mpz_init(x[i])
            mpz_set(x[i], self.rows[i][self.perm[0]])
            mpz_init(y[i])
            mpz_set(y[i], self.rows[i][self.perm[self.dim]])

        j = 1
        for i in range(1, self.dim):
            while j < self.dim and mpz_vec_any_smaller(y, x, self.length):
                for k in range(self.length):
                    mpz_add(y[k], y[k], self.rows[k][self.perm[self.dim+j]])
                j += 1
            if mpz_vec_equal(x, y, self.length):
                ans = True
                break
            for k in range(self.length):
                mpz_add(x[k], x[k], self.rows[k][self.perm[i]])
        else:
            ans = False

        for i in range(2 * self.length):
            mpz_clear(x[i])
        sig_free(x)

        return ans

    def children(self):
        r"""
        Compute the children of the pair ``(p, C)`` composed of a permutation and a
        slice of a Rauzy simplex.

        EXAMPLES::

            sage: from surface_dynamics import *

            sage: p = iet.Permutation('a b c d', 'd c b a')
            sage: F = iet.IETFamily(p, Polyhedron(rays=(ZZ**4).basis())) # optional - pplpy
            sage: top, bot = F.children() # optional - pplpy
            sage: top[0] # optional - pplpy
            't'
            sage: top[1] # optional - pplpy
            top 0 1 2 3
            bot 3 0 2 1
            0 0 0 1
            0 0 1 0
            0 1 0 0
            1 0 0 0
            sage: bot[0] # optional - pplpy
            'b'
            sage: bot[1] # optional - pplpy
            top 0 3 1 2
            bot 3 2 1 0
            0 0 0 1
            0 0 1 0
            0 1 0 0
            1 0 0 0

            sage: p = iet.Permutation('a b c d e f', 'e b f c a d')
            sage: F = iet.IETFamily(p, Polyhedron(rays=[(1,1,1,1,1,2),(0,1,0,1,0,1)])) # optional - pplpy
            sage: c = F.children() # optional - pplpy
            sage: len(c) # optional - pplpy
            1
            sage: c[0][0] # optional - pplpy
            't'
            sage: c[0][1] # optional - pplpy
            top 0 1 2 3 4 5
            bot 4 1 5 3 2 0
            0 1 0 1 0 0
            1 1 1 1 1 1

            sage: F = iet.IETFamily(p, Polyhedron(rays=[(0,0,0,1,0,0),(1,0,0,0,0,0),(1,1,1,1,1,1)])) # optional - pplpy
            sage: c = F.children() # optional - pplpy
            sage: len(c) # optional - pplpy
            1
            sage: c[0][0] # optional - pplpy
            'b'
            sage: c[0][1] # optional - pplpy
            top 0 1 2 3 5 4
            bot 4 1 5 2 0 3
            0 0 0 1 0 0
            1 0 0 0 0 0
            1 1 1 0 1 1
        """
        cdef int itop = self.perm[self.dim - 1]
        cdef int ibot = self.perm[2*self.dim - 1]
        cdef list ans = []
        cdef size_t d = self.poly.affine_dimension()
        cdef IETFamily FF

        cdef C_Polyhedron Ctop = C_Polyhedron(self.poly)
        cdef C_Polyhedron ineq = C_Polyhedron(Variable(itop) >= Variable(ibot))
        ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
        Ctop.intersection_assign(ineq)
        if Ctop.affine_dimension() == d:
            Ctop.affine_image(Variable(itop), Variable(itop) - Variable(ibot))
            FF = IETFamily.__new__(IETFamily)
            FF.dim = self.dim
            FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
            memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
            rauzy_top(FF.perm, FF.dim)
            FF.fill_rays(Ctop.thisptr.minimized_generators())
            FF.poly = Ctop
            ans.append(('t', FF))

        cdef C_Polyhedron Cbot = C_Polyhedron(self.poly)
        ineq = C_Polyhedron(Variable(itop) <= Variable(ibot))
        ineq.add_space_dimensions_and_embed(self.dim - max(ibot, itop) - 1)
        Cbot.intersection_assign(ineq)
        if Cbot.affine_dimension() == d:
            Cbot.affine_image(Variable(ibot), Variable(ibot) - Variable(itop))
            FF = IETFamily.__new__(IETFamily)
            FF.dim = self.dim
            FF.perm = <int *> check_malloc(2 * self.dim * sizeof(int))
            memcpy(FF.perm, self.perm, 2 * sizeof(int) * FF.dim)
            rauzy_bot(FF.perm, FF.dim)
            FF.fill_rays(Cbot.thisptr.minimized_generators())
            FF.poly = Cbot
            ans.append(('b', FF))

        return ans

    def tree(self, max_depth=5, verbose=False):
        r"""
        The tree should be smarter in several ways:

        - once a saddle connection is found it should check whether it is stable and
          whether it corresponds to a splitting, a cylinder decomposition, etc

        - implement a C version!

        - understand what the parabolic are doing!

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: p = iet.Permutation([0,1,2,3,4,5],[5,4,3,2,1,0])
            sage: rays = [[5, 1, 0, 0, 3, 8], [2, 1, 0, 3, 0, 5], [1, 0, 1, 2, 0, 3], [3, 0, 1, 0, 2, 5]]
            sage: F = iet.IETFamily(p, rays) # optional - pplpy
        """
        s = ''
        f = self
        branch = [[(s, f)]]
        seen = set([f])
        while True:
            if verbose:
                print("branch:")
                for ss,ff in branch[-1]:
                    print(ss)
                    print(ff)
                    print()
            while len(branch) < max_depth:
                branch.append([])
                for ss, ff in f.children():
                    if verbose:
                        print("looking children", ss)
                        print("cone", ff)
                    # check for saddles among the children
                    if ff.has_connection():
                        yield 'saddle', s+ss, ff

                    # check for auto-simlarity
                    elif ff in seen:
                        yield 'autosim', s+ss, ff

                    else:
                        branch[-1].append((s+ss, ff))

                if not branch[-1]:
                    branch.pop(-1)
                    break
                else:
                    s, f = branch[-1][-1]
                    assert f not in seen
                    seen.add(f)

            # backtrack
            while branch and len(branch[-1]) == 1:
                s, f = branch.pop()[0]
                if f not in seen:
                    raise RuntimeError("s = {}\nf =\n{}".format(s, f))
                seen.remove(f)

            if not branch:
                return

            s, f = branch[-1].pop(-1)
            seen.remove(f)
            s, f = branch[-1][-1]
            seen.add(f)

    def random_element(self, ring, *args, **kwds):
        r"""
        Return a random point in the cone

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: from surface_dynamics.misc.linalg import deformation_space
            sage: q = iet.Permutation([0,1,2,3,4,5],[5,3,2,1,0,4])
            sage: rays = [[0, 0, 0, 1, 1, 0], [3, 1, 0, 1, 0, 2], [5, 0, 1, 2, 0, 3]]
            sage: F = iet.IETFamily(q, rays) # optional - pplpy
            sage: F # optional - pplpy
            top 0 1 2 3 4 5
            bot 5 3 2 1 0 4
            0 0 0 1 1 0
            3 1 0 1 0 2
            5 0 1 2 0 3
            sage: x = polygen(ZZ)
            sage: K.<a> = NumberField(x^10 - 3, embedding=AA(3)**(1/10))
            sage: T = F.random_element(K) # optional - pplpy
            sage: V = deformation_space(T.lengths()) # optional - pplpy
            sage: (QQ**6).subspace(F.rays()) == V # optional - pplpy
            True
        """
        from constructors import Permutation, IET

        top = [self.perm[i] for i in range(self.dim)]
        bot = [self.perm[i] for i in range(self.dim,2*self.dim)]
        p = Permutation(top, bot, alphabet=range(self.dim))

        coeffs = []
        for _ in range(self.length):
            c = ring.random_element(*args, **kwds)
            while c.is_zero():
                c = ring.random_element(*args, **kwds)
            coeffs.append(c.abs())
        lengths = [sum(coeffs[i] * self.ray_coefficient(i, j) for i in range(self.length)) for j in range(self.dim)]

        return IET(p, lengths)

    # TODO: find a better name and output!!!
    def random_element_statistics(self, ring, num_exp=100, num_iterations=200, *args, **kwds):
        r"""
        INPUT:

        - ``ring`` -- the ring in which is taken the lengths (usually a number field)

        - ``num_exp`` -- number of experiments

        - ``num_iterations`` -- the number of Zorich move to be performed on the
          interval exchange transformations to find a saddle connection

        EXAMPLES::

            sage: from surface_dynamics import *
            sage: q = iet.Permutation([0,1,2,3,4,5],[5,3,2,1,0,4])
            sage: rays = [[0, 0, 0, 1, 1, 0], [3, 1, 0, 1, 0, 2], [5, 0, 1, 2, 0, 3]]
            sage: F = iet.IETFamily(q, rays)  # optional - pplpy
            sage: F.random_element_statistics(K)  # optional - pplpy
            100 saddles found on 100 random point

        A conjectural counterexample to Dynnikov-Skripshenko conjecture::

            sage: path = 'bbbbbtbttbttbbttbtttbbbbttbttbtbbtbbtbbtt'
            sage: p = iet.Permutation('a b c d e f g h i', 'i h g f e d c b a')
            sage: R = p.rauzy_diagram()
            sage: T = R.path(p, *path).self_similar_iet()
            sage: T.sah_arnoux_fathi_invariant()
            (0, 0, 0, 0, 0, 0)
            sage: F = iet.IETFamily(T)  # optional - pplpy
            sage: F  # optional - pplpy
            top 0 1 2 3 4 5 6 7 8
            bot 8 7 6 5 4 3 2 1 0
            71 67 70 107 0 0 19 2 0
            95 63 21 44 35 0 0 0 43
            105 94 88 147 0 0 0 15 19
            253 310 0 243 34 101 68 0 0
            425 395 404 611 14 0 129 0 0
            524 703 0 628 0 245 51 68 0
            619 462 0 229 238 63 0 0 272
            701 910 0 823 0 308 0 119 51
            sage: K.<cbrt3> = NumberField(x^3 - 3, embedding=AA(3)**(1/3))
            sage: F.random_element_statistics(K, num_exp=20, num_iterations=200)  # optional - pplpy # random
            ... saddles found on 20 random points
        """
        num_saddles = 0
        bads = []
        for i in range(num_exp):
            T = self.random_element(ring, *args, **kwds)
            try:
                T.zorich_move(iterations=num_iterations)
            except ValueError:
                num_saddles += 1
            else:
                bads.append(T)
        print('%d saddles found on %d random points' % (num_saddles, num_exp))

    def random_integer_point_statistics(self, num_exp=100):
        r"""
        OUTPUT: a dictionary whose keys are integers ``i`` and values are the
        number of integer points found so that the number of cylinders is
        ``i``.
        """
        # TODO: make iet_interface part of flatsurf
        import iet_interface
        from sage.rings.integer_ring import ZZ
        top, bot = self.permutation()
        dim = len(top)
        cdef dict cyls = {}
        for _ in range(num_exp):
            lengths = sum(ZZ.random_element(0,100000000) * r for r in self.rays())
            ncyls = iet_interface.number_of_cylinders(top, bot, lengths)
            if ncyls not in cyls:
                cyls[ncyls] = 0
            cyls[ncyls] += 1
        return cyls




def tikz_string_to_standalone(filename, s):
    f = open(filename, 'w')
    f.write('\\documentclass{standalone}\n')
    f.write('\\usepackage{tikz}\n')
    f.write('\\begin{document}%\n')
    f.write('\\begin{tikzpicture}\n')
    f.write(s)
    f.write('\\end{tikzpicture}%\n')
    f.write('\\end{document}\n')
    f.close()



r"""
Testing abelian strata and related combinatorics
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import print_function, absolute_import
from six.moves import range, map, filter, zip

import sage.misc.prandom as prandom
from sage.misc.misc import cputime

import random

from sage.combinat.permutation import Permutations

from copy import copy

from sage.groups.perm_gps.permgroup_named import SymmetricGroup
from sage.groups.perm_gps.permgroup import PermutationGroup
from sage.combinat.partition import Partitions

from separatrix_diagram import canonical_perm,SeparatrixDiagram,CylinderDiagram

class Test:
    r"""
    Testing class for ribbon graphs.
    """
    def __init__(self, min_num_seps=5, max_num_seps=10):
        r"""
        Create a ribbon graph testing object.

        INPUT:

        - ``num_darts`` - the number of darts of the randomly generated ribbon
          graphs.
        """
        self.min_num_seps=min_num_seps
        self.max_num_seps=max_num_seps

    def __repr__(self):
        r"""
        Return the string representation of self
        """
        return "Ribbon graph testing class"

    def _do(self, name):
        """
        Perform the test 'test_name', where name is specified as an
        argument. This function exists to avoid a call to eval.
        """

        print("test_%s" % name)
        Test.__dict__["test_%s"%name](self)

    def _get_random_ribbon_graph(self):
        r"""
        Return a random ribbon graph with right parameters.
        """
        n = random.randint(self.min_num_seps,self.max_num_seps)
        S = SymmetricGroup(2*n)

        e = S([(2*i+1,2*i+2) for i in range(n)])
        f = S.random_element()
        P = PermutationGroup([e,f])

        while not P.is_transitive():
            f = S.random_element()
            P = PermutationGroup([e,f])

        return RibbonGraph(
                 edges=[e(i+1)-1 for i in range(2*n)],
                 faces=[f(i+1)-1 for i in range(2*n)])

    def _get_random_cylinder_diagram(self):
        r"""
        Return a random cylinder diagram with right parameters
        """
        test = False
        while test:
            n = random.randint(self.min_num_seps,self.max_num_seps)
            S = SymmetricGroup(2*n)

            bot = S.random_element()
            b = [[i-1 for i in c] for c in bot.cycle_tuples(singletons=True)]

            p = Partitions(2*n,length=len(b)).random_element()
            top = S([i+1 for i in canonical_perm(p)])
            t = [[i-1 for i in c] for c in top.cycle_tuples(singletons=True)]
            prandom.shuffle(t)

            c = CylinderDiagram(zip(b,t))
            test = c.is_connected()

        return c

    def random(self, seconds=0):
        """
        Perform random tests for a given number of seconds, or
        indefinitely if seconds is not specified.
        """
        self.test("random", seconds)

    def test(self, name, seconds=0):
        """
        Repeatedly run 'test_name', where name is passed as an
        argument. If seconds is nonzero, run for that many seconds. If
        seconds is 0, run indefinitely.
        """
        seconds = float(seconds)
        total = cputime()
        n = 1
        while seconds == 0 or cputime(total) < seconds:
            s = "** test_dimension: number %s"%n
            if seconds > 0:
                s += " (will stop after about %s seconds)"%seconds
            t = cputime()
            self._do(name)
            print("\ttime=%s\telapsed=%s" % (cputime(t), cputime(total)))
            n += 1

    def test_random(self):
        """
        Do a random test from all the possible tests.

        EXAMPLES::

            sage: from sage.modular.arithgroup.tests import Test
            sage: Test().test_random() #random
            Doing random test
        """
        tests = [a for a in Test.__dict__.keys() if a[:5] == "test_" and a != "test_random"]
        name = prandom.choice(tests)
        print("Doing random test %s" % name)
        Test.__dict__[name](self)

    def test_chain_complex(self):
        r"""
        Test dimensions of chain complex as well as euler characteristic and
        rank of the intersection form.
        """
        g = self._get_random_ribbon_graph()
        C = g.chain_complex()

        assert C.chain_space(0).rank() == g.num_vertices()
        assert C.chain_space(1).rank() == g.num_edges()
        assert C.chain_space(2).rank() == g.num_faces()

        assert C.cycle_space(0).rank() - C.boundary_space(0).rank() == 0  # this is reduced homology
        assert C.cycle_space(1).rank() - C.boundary_space(1).rank() == 2*g.genus()
        assert C.cycle_space(2).rank() - C.boundary_space(2).rank() == 1

        c,m = g.cycle_basis(intersection_matrix=True)
        assert m.is_skew_symmetric()
        assert m.rank() == 2*g.genus()

    def test_homological_dimension(self):
        from sage.rings.finite_rings.finite_field_constructor import GF

        g = self._get_random_cylinder_diagram()
        C = g.to_ribbon_graph().chain_complex(GF(2))
        Z1 = C.cycle_space(1)
        B1 = C.boundary_space(1)
        V = Z1.submodule(g.circumferences_of_cylinders(GF(2)))

        assert g.homological_dimension_of_cylinders() ==  V.rank() - V.intersection(B1).rank()



r"""
Flat surfaces
"""
#*****************************************************************************
#       Copyright (C) 2019 Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
#*****************************************************************************

from __future__ import absolute_import

from .version import version as __version__

import warnings
warnings.filterwarnings('default',
    r'\[surface_dynamics].*')

# Make sure that sage's imports are going to resolve in the correct order
import sage.all

from .flat_surfaces.all import *
from .interval_exchanges.all import *
from .misc.constellation import Constellation, Constellations
from .topological_recursion import *
from .topology.all import *

del absolute_import
